#!/user/bin/env python3
# coding: utf-8


from copy import deepcopy


def recursive_sort(structure, key=None):
    if isinstance(structure, (list, tuple, set, frozenset)):
        return [recursive_sort(x, key) for x in sorted(structure, key=key)]
    elif isinstance(structure, dict):
        return {k: recursive_sort(v, key) for k, v in structure.items()}
    return deepcopy(structure)





#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import itertools
from collections import deque
from itertools import chain, combinations

from six import moves as six_moves


def chunkwize(chunksize, iterable):
    """
    >>> list(chunkwize(5, range(14)))
    [[0, 1, 2, 3, 4], 
     [5, 6, 7, 8, 9], 
     [10, 11, 12, 13]]
     
    :param chunksize: integer
    :param iterable:
    """
    chunksize = int(chunksize)
    chunk = []
    for item in iterable:
        if len(chunk) >= chunksize:
            yield chunk
            chunk = []
        chunk.append(item)
    yield chunk


def chunkwize_parallel(chunksize, *sequences):
    """
    >>> s1 = '1234567890'
    >>> s2 = 'abcdefghijk'
    >>> list(chunkwize_parallel(4, s1, s2))
    [['1234', 'abcd'], ['5678', 'efgh'], ['90', 'ijk']]
    
    :param chunksize: integer
    :param sequences: tuple of strings or lists (must support slicing!)
    """
    chunksize = int(chunksize)
    for i in itertools.count(0):
        r = [s[i * chunksize:(i + 1) * chunksize] for s in sequences]
        if any(r):
            yield r
        else:
            raise StopIteration


def all_combinations(iterable):
    """
    >>> list(all_combinations('abcd'))
    [(),
     ('a',),
     ('b',),
     ('c',),
     ('d',),
     ('a', 'b'),
     ('a', 'c'),
     ('a', 'd'),
     ('b', 'c'),
     ('b', 'd'),
     ('c', 'd'),
     ('a', 'b', 'c'),
     ('a', 'b', 'd'),
     ('a', 'c', 'd'),
     ('b', 'c', 'd'),
     ('a', 'b', 'c', 'd')]
    """
    items = list(iterable)
    return chain.from_iterable(
        combinations(items, i) for i in range(1 + len(items)))


def window_sum(wsize, numbers):
    """
    >>> numbers = [1, 10, 100, 1000, 10000, 100000]
    >>> list(window_sum(3, numbers))
    [111, 1110, 11100, 111000]
    
    :param wsize: integer, size of the moving window
    :param numbers: an iterable of numbers
    """
    queue = deque(maxlen=wsize)
    for num in numbers:
        queue.append(num)
        if len(queue) >= wsize:
            yield sum(queue)


_void = object()


def alternate(*iterables, fill=_void):
    """
    >>> ''.join(list(alternate('ABCD', 'abcde')))
    'AaBbCcDde'
    >>> ''.join(list(alternate('ABCD', 'abcde', fill='_')))
    'AaBbCcDd_e' 
    """
    zip_longest = six_moves.zip_longest
    alt = itertools.chain(*zip_longest(*iterables, fillvalue=fill))
    for item in alt:
        if item is not _void:
            yield item





#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function
import math


def floor(number):
    return int(math.floor(number))


def ceil(number):
    return int(math.ceil(number))


def overlap_size(va, vb, ua, ub):
    """
    Calculate overlap size of intervals (va, vb) and (ua, ub)
    """
    alen = abs(va - vb)
    blen = abs(ua - ub)
    wide = max(va, vb, ua, ub) - min(va, vb, ua, ub)
    return alen + blen - wide



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import os
import sys


def under_homedir(*paths):
    if sys.platform == 'win32':
        homedir = os.environ["HOMEPATH"]
    else:
        homedir = os.path.expanduser('~')
    return os.path.join(homedir, *paths)



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import socket


def validate_ipv4_address(address):
    # http://stackoverflow.com/a/4017219/2925169
    try:
        socket.inet_pton(socket.AF_INET, address)
    except AttributeError:  # no inet_pton here, sorry
        try:
            socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:  # not a valid address
        return False
    return True


def validate_ipv6_address(address):
    # http://stackoverflow.com/a/4017219/2925169
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:  # not a valid address
        return False
    return True



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import itertools

import six
from joker.cast.numeric import ceil


class _ListWrapper(object):
    def __init__(self, iterable):
        self._items = list(iterable)

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({})'.format(c, self._items)


class Pool(_ListWrapper):
    def __init__(self, iterable):
        super(Pool, self).__init__(iterable)
        self._cycle = itertools.cycle(self._items)

    def shuffle(self):
        import random
        random.shuffle(self._items)
        self._cycle = itertools.cycle(self._items)
        return self

    def take(self, n):
        return list(itertools.islice(self._cycle, n))


class Circular(_ListWrapper):
    """
    >>> c = Circular([0, 1, 2, 3, 4]) 
    >>> list(c[-1:5])
    [4, 0, 1, 2, 3, 4]
    """
    def __init__(self, iterable):
        super(Circular, self).__init__(iterable)
        assert self._items

    def ix_turn(self, ix):
        if ix is None:
            return
        return len(self._items) - 1 - ix

    def ix_shift(self, *indexes):
        m = min(i for i in indexes if i is not None)
        if m >= 0:
            return indexes
        n = len(self._items)
        shift = ceil(-1. * m / n) * n
        new_indexes = []
        for ix in indexes:
            if ix is None:
                new_indexes.append(None)
            else:
                new_indexes.append(ix + shift)
        return tuple(new_indexes)

    def standardize(self, slc):
        """
        :param slc: a slice instance 
        :return: (start, stop, step)  # a tuple  
        """
        assert isinstance(slc, slice)
        if slc.step is None or slc.step >= 0:
            return self.ix_shift(slc.start, slc.stop, slc.step)
        return self.ix_shift(
            self.ix_turn(slc.start),
            self.ix_turn(slc.stop),
            -slc.step)

    def __getitem__(self, key):
        n = len(self._items)
        if isinstance(key, six.integer_types):
            return self._items[key % n]

        if isinstance(key, slice):
            start, stop, step = self.standardize(key)
            if key.step and key.step < 0:
                c_items = itertools.cycle(self._items[::-1])
            else:
                c_items = itertools.cycle(self._items)
            return itertools.islice(c_items, start, stop, step)


class CircularString(object):
    """
    >>> cs = CircularString('0123456789')
    >>> cs[-1:12]
    '9012345678901'
    """
    def __init__(self, string):
        self._string = string
        self._circular = Circular(string)

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({})'.format(c, self._string)

    def __getitem__(self, key):
        return ''.join(list(self._circular[key]))





#!/user/bin/env python3
# coding: utf-8
from copy import deepcopy
import six


tail_for_str = '... +{} trailing characters'
tail_for_arr = '... +{} trailing objects'


def _str_trans(s):
    return s.replace('\n', r'\n').replace('\r', r'\r')


# TODO: test this func
def minify(obj, inplace=False, listlimit=1, strlimit=200):
    """
    Shorten list and strings to reduce verbosity.
    :param listlimit: 
    :param strlimit: 
    :param obj: input structure (dict or list)
    :param inplace: do minification on `obj` or a copy of it
    :return: a structure similar to `structure`(dict or list)
    """

    if not inplace:
        obj = deepcopy(obj)

    if isinstance(obj, list):
        # remove items beyond listlimit
        if len(obj) > listlimit:
            tail = tail_for_arr.format(len(obj)-listlimit)
            obj[listlimit:] = []
        else:
            tail = None

        for x in obj:
            minify(x, True, listlimit, strlimit)

        # append trail if trailing items have been removed
        if tail is not None:
            obj.append(tail)

    elif isinstance(obj, six.string_types):
        if isinstance(obj, six.binary_type):
            obj = obj.decode('utf-8')
        if len(obj) > strlimit:
            tail = tail_for_str.format(len(obj)-strlimit)
            obj = obj[:strlimit] + tail

    elif isinstance(obj, dict):
        for key, val in obj.items():
            val_ = minify(val, True, listlimit, strlimit)
            obj[key] = val_

    return obj


# TODO: test this func
def diff(cat, dog, inplace=False):
    """
    Compare 2 dicts recursively. Cancel out shared keys corresponding to
    values of same structure (recursively type match).
    :param cat: A dict
    :param dog: Another dict
    :param inplace: Cancel out similar fields in place, default False
    :return: (cat, dog) after cancellation
    """
    if not inplace:
        cat = deepcopy(cat)
        dog = deepcopy(dog)

    noncontainer_types = int, six.text_type, six.binary_type, float, type(None)

    cat_keys = set(cat.keys())
    dog_keys = set(dog.keys())
    shared_keys = set.intersection(cat_keys, dog_keys)
    for key in shared_keys:
        cat_val = cat[key]
        dog_val = dog[key]

        if type(cat_val) == type(dog_val):

            # 2 dicts: recursively diff
            if isinstance(cat_val, dict):
                diff(cat_val, dog_val, True)
                # del this key only if both become {}
                if (not cat_val) and (not dog_val):
                    del cat[key]
                    del dog[key]

            # 2 lists: are they empty?
            elif isinstance(cat_val, list):
                # 2 empty lists
                if not (cat_val or dog_val):
                    del cat[key]
                    del dog[key]

                # 2 non-empty lists
                elif cat_val and dog_val:
                    # assuming all lists are homogeneous
                    # TODO: check if the lists are homogeneous
                    cat[key] = [cat_val[0]]
                    dog[key] = [dog_val[0]]
                    cat_ = {'v': cat_val[0]}
                    dog_ = {'v': dog_val[0]}
                    diff(cat_, dog_, True)
                    if (not cat_val[0]) and (not dog_val[0]):
                        del cat[key]
                        del dog[key]

            elif isinstance(cat_val, noncontainer_types):
                del cat[key]
                del dog[key]

    return bool(cat and dog), cat, dog



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import math
import random

import numpy


class Bitmap(object):
    def __init__(self, size, bitwidth=32, array=None, members=None):
        """
        :param size: number of bits
        :param bitwidth:
        :param array:
        :param members:
        :return:
        """
        self.size = int(math.ceil(size))
        self.bitwidth = bitwidth
        if array is None:
            dtype = 'uint{}'.format(bitwidth)
            arrsize = int(numpy.ceil(size * 1. / bitwidth))
            self.array = numpy.zeros(arrsize, dtype=dtype)
        else:
            self.array = array
        if members:
            self.update(members)

    def _verify(self, bitmap):
        if self.size != bitmap.size:
            raise ValueError('cannot use a bitmap of different size')
        if self.bitwidth != bitmap.bitwidth:
            raise ValueError('cannot use a bitmap of different bitwidth')

    def add(self, integer):
        idx = int(integer / self.bitwidth)
        self.array[idx] |= 1 << (integer % self.bitwidth)

    def __contains__(self, integer):
        idx = int(integer / self.bitwidth)
        return self.array[idx] & (1 << (integer % self.bitwidth))

    def update(self, collection):
        if isinstance(collection, Bitmap):
            self._verify(collection)
            self.array |= collection.array
            return
        for integer in collection:
            self.add(integer)

    def union(self, collction):
        this = self.copy()
        this.update(collction)
        return this

    def intersection_update(self, collection):
        if not isinstance(collection, Bitmap):
            that = self.metacopy()
            that.update(collection)
        else:
            that = collection
        self.array &= that.array

    def intersection(self, collection):
        this = self.copy()
        this.intersection_update(collection)
        return this

    def difference_update(self, collection):
        if not isinstance(collection, Bitmap):
            that = self.metacopy()
            that.update(collection)
        else:
            that = collection
        self.intersection_update(-that)

    def difference(self, collection):
        this = self.copy()
        this.difference_update(collection)
        return this

    def clear(self):
        self.array[:] = 0
        return self

    def copy(self):
        array = self.array.copy()
        return Bitmap(self.size, self.bitwidth, array=array)

    def metacopy(self):
        array = numpy.zeros_like(self.array)
        return Bitmap(self.size, self.bitwidth, array=array)

    def __neg__(self):
        array = ~ self.array
        return Bitmap(self.size, self.bitwidth, array=array)

    def __iter__(self):
        for plane in self._iter_bit_planes(False):
            for i in plane:
                yield i

    def get_members(self, count=None, random_start=False):
        members = []
        for plane in self._iter_bit_planes(random_start):
            members.extend(plane)
            if len(members) >= count:
                break
        return members

    def pick(self):
        for plane in self._iter_bit_planes(random_start=True):
            members = plane.tolist()
            if members:
                return random.choice(members)

    def _iter_bit_planes(self, random_start):
        start = random.randrange(self.bitwidth) if random_start else 0
        for j in [(i + start) % self.bitwidth for i in range(self.bitwidth)]:
            for arr in numpy.nonzero(self.array & numpy.array([1 << j])):
                # arr: indexes with bit 1 on j-th bit
                yield numpy.array([j]) + arr * self.bitwidth

    @classmethod
    def test_bitmap(cls):
        bm = cls(1001)
        items = [1, 34, 555, 90]
        for i in range(1001):
            assert i not in bm
        bm.update(items)
        for i in range(1001):
            if i in items:
                assert i in bm
            else:
                assert i not in bm
        return bm
